import os
import sys
import json
from pathlib import Path
from datetime import datetime

try:
    import requests
    import yaml
except ImportError as e:
    print(f"Missing dependency: {e}")
    print("Install with: pip install requests pyyaml")
    sys.exit(1)


SKYLOS_DIR = ".skylos"
CONFIG_FILE = "config.yaml"
SUPPRESSIONS_FILE = "suppressions.json"
META_FILE = ".sync-meta.json"
DEFAULT_API_URL = "https://skylos.dev"

GLOBAL_CREDS_DIR = Path.home() / ".skylos"
GLOBAL_CREDS_FILE = GLOBAL_CREDS_DIR / "credentials.json"


def get_api_url():
    return os.environ.get("SKYLOS_API_URL", DEFAULT_API_URL)


def get_token():
    env_token = os.environ.get("SKYLOS_TOKEN", "").strip()
    if env_token:
        return env_token

    if GLOBAL_CREDS_FILE.exists():
        try:
            data = json.loads(GLOBAL_CREDS_FILE.read_text())
            return data.get("token")
        except:
            pass

    return None


def save_token(token, project_name=None, org_name=None, plan=None):
    GLOBAL_CREDS_DIR.mkdir(exist_ok=True)

    data = {
        "token": token,
        "saved_at": datetime.utcnow().isoformat() + "Z",
    }
    if project_name:
        data["project_name"] = project_name
    if org_name:
        data["org_name"] = org_name
    if plan:
        data["plan"] = plan

    GLOBAL_CREDS_FILE.write_text(json.dumps(data, indent=2))
    return str(GLOBAL_CREDS_FILE)


def clear_token():
    if GLOBAL_CREDS_FILE.exists():
        GLOBAL_CREDS_FILE.unlink()
        return True
    return False


def mask_token(token):
    if not token or len(token) <= 12:
        return "****"
    return token[:8] + "..." + token[-4:]


class AuthError(Exception):
    pass


def api_get(endpoint, token):
    url = f"{get_api_url()}{endpoint}"

    try:
        resp = requests.get(
            url,
            headers={"Authorization": f"Bearer {token}"},
            timeout=30,
        )
    except requests.exceptions.ConnectionError:
        raise AuthError(f"Cannot connect to {get_api_url()}")
    except requests.exceptions.Timeout:
        raise AuthError("Request timed out")

    if resp.status_code == 401:
        raise AuthError("Invalid API token")

    resp.raise_for_status()
    return resp.json()


def cmd_connect(token_arg=None):
    print("\n Connect to Skylos Cloud\n")

    token = token_arg or os.environ.get("SKYLOS_TOKEN", "").strip()

    if not token:
        print("Enter your API token (from Dashboard → Settings):")
        try:
            token = input("> ").strip()
        except (KeyboardInterrupt, EOFError):
            print("\nCancelled.")
            sys.exit(1)

    if not token:
        print("Error: No token provided.")
        sys.exit(1)

    print(f"Verifying token {mask_token(token)}...")

    try:
        info = api_get("/api/sync/whoami", token)
    except AuthError as e:
        print(f"\n✗ {e}")
        sys.exit(1)

    project = info.get("project", {})
    org = info.get("organization", {})
    plan = info.get("plan", "free")

    print(f"\n✓ Connected!\n")
    print(f"  Project:      {project.get('name', 'Unknown')}")
    print(f"  Organization: {org.get('name', 'Unknown')}")
    print(f"  Plan:         {plan.capitalize()}")

    creds_path = save_token(
        token, project_name=project.get("name"), org_name=org.get("name"), plan=plan
    )

    print(f"\nToken saved to {creds_path}")
    print("\nYou can now run:")
    print("  skylos .           # Scan locally")
    print("  skylos . --upload  # Scan and upload")


def cmd_status():
    token = get_token()

    if not token:
        print("\nNot connected to Skylos Cloud.")
        print("Run 'skylos sync connect' to connect.\n")
        return

    print(f"\nChecking connection...")

    try:
        info = api_get("/api/sync/whoami", token)
    except AuthError as e:
        print(f"\n✗ {e}")
        print("Run 'skylos sync connect' to reconnect.\n")
        return

    project = info.get("project", {})
    org = info.get("organization", {})
    plan = info.get("plan", "free")

    print(f"\n✓ Connected\n")
    print(f"  Project:      {project.get('name', 'Unknown')}")
    print(f"  Organization: {org.get('name', 'Unknown')}")
    print(f"  Plan:         {plan.capitalize()}")


def cmd_disconnect():
    if clear_token():
        print("✓ Disconnected.")
    else:
        print("No saved credentials found.")


def cmd_pull():
    token = get_token()

    if not token:
        print("Error: Not connected.")
        print("Run 'skylos sync connect' first.")
        sys.exit(1)

    skylos_dir = Path(SKYLOS_DIR)
    skylos_dir.mkdir(exist_ok=True)

    try:
        info = api_get("/api/sync/whoami", token)
        print(f"Connected to: {info.get('project', {}).get('name', 'Unknown')}\n")
    except AuthError as e:
        print(f"Error: {e}")
        sys.exit(1)

    try:
        print("Pulling configuration...")
        config_data = api_get("/api/sync/config", token)

        config_path = skylos_dir / CONFIG_FILE
        with config_path.open("w") as f:
            yaml.dump(config_data.get("config", {}), f, default_flow_style=False)
        print(f"  ✓ {config_path}")

        print("Pulling suppressions...")
        supp_data = api_get("/api/sync/suppressions", token)

        supp_path = skylos_dir / SUPPRESSIONS_FILE
        with supp_path.open("w") as f:
            json.dump(supp_data.get("suppressions", []), f, indent=2)
        print(f"  ✓ {supp_path} ({supp_data.get('count', 0)} suppressions)")

        print("\n✓ Sync complete!")

    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    if not args:
        print("Usage: skylos sync <command>")
        print("")
        print("Commands:")
        print("  connect [token]  Connect to Skylos Cloud")
        print("  status           Show connection status")
        print("  disconnect       Remove saved credentials")
        print("  pull             Pull config and suppressions")
        return

    cmd = args[0].lower()

    if cmd == "connect":
        cmd_connect(args[1] if len(args) > 1 else None)
    elif cmd == "status":
        cmd_status()
    elif cmd == "disconnect":
        cmd_disconnect()
    elif cmd == "pull":
        cmd_pull()
    else:
        print(f"Unknown command: {cmd}")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import ast
import sys
import json
import logging
import os
from pathlib import Path
from collections import defaultdict
from skylos.visitor import Visitor
from skylos.constants import AUTO_CALLED
from skylos.visitors.test_aware import TestAwareVisitor
from skylos.rules.secrets import scan_ctx as _secrets_scan_ctx
from skylos.rules.danger.danger import scan_ctx as scan_danger
import os
import traceback
from skylos.visitors.framework_aware import (
    FrameworkAwareVisitor
)
from skylos.config import get_all_ignore_lines, load_config
from skylos.visitors.languages.typescript import scan_typescript_file

from skylos.linter import LinterVisitor
from skylos.rules.quality.complexity import ComplexityRule
from skylos.rules.quality.nesting import NestingRule
from skylos.rules.danger.calls import DangerousCallsRule
from skylos.rules.quality.structure import ArgCountRule, FunctionLengthRule
from skylos.rules.quality.logic import (
    MutableDefaultRule,
    BareExceptRule,
    DangerousComparisonRule,
    TryBlockPatternsRule,
)
from skylos.penalties import apply_penalties
from skylos.rules.quality.performance import PerformanceRule
from skylos.rules.quality.unreachable import UnreachableCodeRule

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger("Skylos")


class Skylos:
    def __init__(self):
        self.defs = {}
        self.refs = []
        self.dynamic = set()
        self.exports = defaultdict(set)

    def _module(self, root, f):
        p = list(f.relative_to(root).parts)
        if p[-1].endswith(".py"):
            p[-1] = p[-1][:-3]
        if p[-1] == "__init__":
            p.pop()
        return ".".join(p)

    def _should_exclude_file(self, file_path, root_path, exclude_folders):
        if not exclude_folders:
            return False

        try:
            rel_path = file_path.relative_to(root_path)
        except ValueError:
            return False

        path_parts = rel_path.parts
        rel_path_str = str(rel_path).replace("\\", "/")

        for exclude_folder in exclude_folders:
            exclude_normalized = exclude_folder.replace("\\", "/")

            if "*" in exclude_folder:
                for part in path_parts:
                    if part.endswith(exclude_folder.replace("*", "")):
                        return True
            elif "/" in exclude_normalized:
                if rel_path_str == exclude_normalized:
                    return True
                if rel_path_str.startswith(exclude_normalized + "/"):
                    return True
                check = "/" + rel_path_str + "/"
                if "/" + exclude_normalized + "/" in check:
                    return True
            else:
                if exclude_folder in path_parts:
                    return True

        return False

    def _get_python_files(self, path, exclude_folders=None):
        p = Path(path).resolve()

        if p.is_file():
            if p.suffix == '.pyi':
                return [], p.parent
            return [p], p.parent

        root = p
        all_files = []
        for f in p.glob("**/*.py"):
            if f.suffix == ".py":
                all_files.append(f)

        if exclude_folders:
            filtered_files = []
            excluded_count = 0

            for file_path in all_files:
                if self._should_exclude_file(file_path, root, exclude_folders):
                    excluded_count += 1
                    continue
                filtered_files.append(file_path)

            if excluded_count > 0:
                logger.info(f"Excluded {excluded_count} files from analysis")

            return filtered_files, root

        return all_files, root

    def _mark_exports(self):
        for name, definition in self.defs.items():
            if definition.in_init and not definition.simple_name.startswith("_"):
                definition.is_exported = True
        
        all_exported_names = set()
        for mod, export_names in self.exports.items():
            all_exported_names.update(export_names)

        for def_name, def_obj in self.defs.items():
            if def_obj.simple_name in all_exported_names:
                def_obj.is_exported = True
                def_obj.references += 1

        for mod, export_names in self.exports.items():
            for name in export_names:
                for def_name, def_obj in self.defs.items():
                    if (
                        def_name.startswith(f"{mod}.")
                        and def_obj.simple_name == name
                        and def_obj.type != "import"
                    ):
                        def_obj.is_exported = True

    def _mark_refs(self, progress_callback=None):
        total_refs = len(self.refs)
        if progress_callback:
            progress_callback(0, total_refs or 1, Path("PHASE: mark refs"))

        import_to_original = {}
        for name, def_obj in self.defs.items():
            if def_obj.type == "import":
                import_name = name.split(".")[-1]

                for def_name, orig_def in self.defs.items():
                    if (
                        orig_def.type != "import"
                        and orig_def.simple_name == import_name
                        and def_name != name
                    ):
                        import_to_original[name] = def_name
                        break

        simple_name_lookup = defaultdict(list)
        for definition in self.defs.values():
            simple_name_lookup[definition.simple_name].append(definition)

        total_refs = len(self.refs)
        tick_every = int(os.getenv("SKYLOS_MARKREFS_TICK", "5000"))

        for i, (ref, ref_file) in enumerate(self.refs, 1):
            if progress_callback and (i == 1 or i % tick_every == 0 or i == total_refs):
                progress_callback(i, total_refs or 1, Path("PHASE: mark refs"))
            file_key = f"{ref_file}:{ref}"

            if file_key in self.defs:
                self.defs[file_key].references += 1
                if file_key in import_to_original:
                    original = import_to_original[file_key]
                    if original in self.defs:
                        self.defs[original].references += 1
                continue

            if ref in self.defs:
                self.defs[ref].references += 1
                if ref in import_to_original:
                    original = import_to_original[ref]
                    self.defs[original].references += 1
                continue

            if "." in ref:
                ref_mod, simple = ref.rsplit(".", 1)
            else:
                ref_mod, simple = "", ref
            candidates = simple_name_lookup.get(simple, [])

            if ref_mod:
                if ref_mod in ("cls", "self"):
                    cls_candidates = []
                    for d in candidates:
                        if d.type == "variable" and "." in d.name:
                            cls_candidates.append(d)

                    if cls_candidates:
                        for d in cls_candidates:
                            d.references += 1
                        continue

                else:
                    filtered = []
                    for d in candidates:
                        if d.name.startswith(ref_mod + ".") and d.type != "import":
                            filtered.append(d)
                    candidates = filtered
            else:
                filtered = []
                for d in candidates:
                    if d.type != "import":
                        filtered.append(d)
                candidates = filtered

            if len(candidates) > 1:
                same_file = []
                for d in candidates:
                    if str(d.filename) == str(ref_file):
                        same_file.append(d)
                if len(same_file) == 1:
                    candidates = same_file

            if len(candidates) == 1:
                candidates[0].references += 1
                continue

            if len(candidates) > 1:
                if not ref_mod or ref_mod in ("self", "cls"):
                    for d in candidates:
                        d.references += 1
                    continue

            non_import_defs = []

            non_import_defs = []
            for d in simple_name_lookup.get(simple, []):
                if d.type != "import":
                    non_import_defs.append(d)

            if len(non_import_defs) == 1:
                non_import_defs[0].references += 1
                continue

            if "." in ref:
                ref_simple = ref.split(".")[-1]
                same_file_methods = []
                for d in self.defs.values():
                    if d.type == "method" and d.simple_name == ref_simple:
                        if str(d.filename) == str(ref_file):
                            same_file_methods.append(d)

                if same_file_methods:
                    for m in same_file_methods:
                        m.references += 1
                    continue

        if hasattr(self, "pattern_trackers"):
            seen_trackers = set()
            for _, tracker in self.pattern_trackers.items():
                if id(tracker) in seen_trackers:
                    continue
                seen_trackers.add(id(tracker))
                for def_obj in self.defs.values():
                    should_mark, _, _ = tracker.should_mark_as_used(def_obj)
                    if should_mark:
                        def_obj.references += 1

        from skylos.implicit_refs import pattern_tracker as global_tracker

        if global_tracker.traced_calls:
            for def_obj in self.defs.values():
                should_mark, _, reason = global_tracker.should_mark_as_used(def_obj)
                if should_mark:
                    def_obj.references += 1

    def _get_base_classes(self, class_name):
        if class_name not in self.defs:
            return []

        class_def = self.defs[class_name]

        if hasattr(class_def, "base_classes"):
            return class_def.base_classes

        return []

    def _apply_heuristics(self):
        class_methods = defaultdict(list)
        for definition in self.defs.values():
            if definition.type in ("method", "function") and "." in definition.name:
                cls = definition.name.rsplit(".", 1)[0]
                if cls in self.defs and self.defs[cls].type == "class":
                    class_methods[cls].append(definition)

        for cls, methods in class_methods.items():
            if self.defs[cls].references > 0:
                for method in methods:
                    if method.simple_name in AUTO_CALLED:
                        method.references += 1

                    if (
                        method.simple_name.startswith("visit_")
                        or method.simple_name.startswith("leave_")
                        or method.simple_name.startswith("transform_")
                    ):
                        method.references += 1

                    if method.simple_name == "format" and cls.endswith("Formatter"):
                        method.references += 1

    def analyze(
        self,
        path,
        thr=60,
        exclude_folders=None,
        enable_secrets=False,
        enable_danger=False,
        enable_quality=False,
        extra_visitors=None,
        progress_callback=None,
    ):
        files, root = self._get_python_files(path, exclude_folders)

        if not files:
            logger.warning(f"No Python files found in {path}")
            return json.dumps(
                {
                    "unused_functions": [],
                    "unused_imports": [],
                    "unused_classes": [],
                    "unused_variables": [],
                    "unused_parameters": [],
                    "unused_files": [],
                    "analysis_summary": {
                        "total_files": 0,
                        "excluded_folders": exclude_folders if exclude_folders else [],
                    },
                }
            )

        logger.info(f"Analyzing {len(files)} Python files...")

        modmap = {}
        for f in files:
            modmap[f] = self._module(root, f)

        from skylos.implicit_refs import pattern_tracker

        from skylos.implicit_refs import pattern_tracker as global_pattern_tracker

        global_pattern_tracker.known_refs.clear()
        global_pattern_tracker._compiled_patterns.clear()
        global_pattern_tracker.f_string_patterns.clear()
        global_pattern_tracker.coverage_hits.clear()
        global_pattern_tracker.covered_files_lines.clear()
        global_pattern_tracker._coverage_by_basename.clear()
        global_pattern_tracker.traced_calls.clear()
        global_pattern_tracker.traced_by_file.clear()
        global_pattern_tracker._traced_by_basename.clear()

        project_root = Path(path).resolve()
        if not project_root.is_dir():
            project_root = project_root.parent

        coverage_path = project_root / ".coverage"
        if coverage_path.exists():
            if global_pattern_tracker.load_coverage():
                logger.info(
                    f"Loaded coverage data ({len(pattern_tracker.coverage_hits)} lines)"
                )

        root = Path(path).resolve()
        if root.is_dir():
            project_root = root
        else:
            project_root = root.parent

        trace_path = project_root / ".skylos_trace"
        if trace_path.exists():
            pattern_tracker.load_trace(str(trace_path))


        all_secrets = []
        all_dangers = []
        all_quality = []
        empty_files = []
        file_contexts = []

        pattern_trackers = {}

        for i, file in enumerate(files):
            if progress_callback:
                progress_callback(i + 1, len(files), file)
            mod = modmap[file]
            (
                defs,
                refs,
                dyn,
                exports,
                test_flags,
                framework_flags,
                q_finds,
                d_finds,
                pro_finds,
                pattern_tracker,
                empty_file_finding,
                cfg,
            ) = proc_file(file, mod, extra_visitors)

            if pattern_tracker:
                pattern_trackers[mod] = pattern_tracker

            for definition in defs:
                if definition.type == "import":
                    key = f"{definition.filename}:{definition.name}"
                else:
                    key = definition.name
                self.defs[key] = definition

            self.refs.extend(refs)
            self.dynamic.update(dyn)
            self.exports[mod].update(exports)

            file_contexts.append((defs, test_flags, framework_flags, file, mod, cfg))

            if empty_file_finding:
                empty_files.append(empty_file_finding)

            if enable_quality and q_finds:
                all_quality.extend(q_finds)

            if enable_danger and d_finds:
                all_dangers.extend(d_finds)

            if pro_finds:
                all_dangers.extend(pro_finds)

            if enable_secrets and _secrets_scan_ctx is not None:
                try:
                    src = Path(file).read_text(encoding="utf-8", errors="ignore")
                    src_lines = src.splitlines(True)
                    rel = str(Path(file).relative_to(root))
                    ctx = {"relpath": rel, "lines": src_lines, "tree": None}
                    findings = list(_secrets_scan_ctx(ctx))
                    if findings:
                        all_secrets.extend(findings)
                except Exception:
                    pass
      
        self.pattern_trackers = pattern_trackers

        self._global_abc_classes = set()
        self._global_protocol_classes = set()
        self._global_abstract_methods = {}
        self._global_abc_implementers = {}
        self._global_protocol_implementers = {}
        self._global_protocol_method_names = {}
        
        for defs, test_flags, framework_flags, file, mod, cfg in file_contexts:
            self._global_abc_classes.update(getattr(framework_flags, "abc_classes", set()))
            self._global_protocol_classes.update(getattr(framework_flags, "protocol_classes", set()))
            
            for cls, methods in getattr(framework_flags, "abstract_methods", {}).items():
                if cls not in self._global_abstract_methods:
                    self._global_abstract_methods[cls] = set()
                self._global_abstract_methods[cls].update(methods)
            
            for cls, parents in getattr(framework_flags, "abc_implementers", {}).items():
                if cls not in self._global_abc_implementers:
                    self._global_abc_implementers[cls] = []
                self._global_abc_implementers[cls].extend(parents)
            
            for cls, parents in getattr(framework_flags, "protocol_implementers", {}).items():
                if cls not in self._global_protocol_implementers:
                    self._global_protocol_implementers[cls] = []
                self._global_protocol_implementers[cls].extend(parents)
            
            for cls, methods in getattr(framework_flags, "protocol_method_names", {}).items():
                if cls not in self._global_protocol_method_names:
                    self._global_protocol_method_names[cls] = set()
                self._global_protocol_method_names[cls].update(methods)
        
        self._duck_typed_implementers = set()

        class_methods = {}
        for def_obj in self.defs.values():
            if def_obj.type == "method" and "." in def_obj.name:
                parts = def_obj.name.split(".")
                if len(parts) >= 2:
                    class_name = parts[-2]
                    method_name = parts[-1]
                    if class_name not in class_methods:
                        class_methods[class_name] = set()
                    class_methods[class_name].add(method_name)
        
        for class_name, methods in class_methods.items():
            if class_name in self._global_protocol_classes:
                continue

            if class_name in self._global_protocol_implementers:
                continue
            
            for protocol_name, protocol_methods in self._global_protocol_method_names.items():
                if not protocol_methods or len(protocol_methods) < 3:
                    continue
                
                matching = methods & protocol_methods
                match_ratio = len(matching) / len(protocol_methods)
                
                if match_ratio >= 0.7 and len(matching) >= 3:
                    self._duck_typed_implementers.add(class_name)
                    break

        for defs, test_flags, framework_flags, file, mod, cfg in file_contexts:
            for definition in defs:
                apply_penalties(self, definition, test_flags, framework_flags, cfg)

            if enable_danger and scan_danger is not None:
                try:
                    findings = scan_danger(root, [file])
                    if findings:
                        all_dangers.extend(findings)
                except Exception as e:
                    logger.error(f"Error scanning {file} for dangerous code: {e}")
                    if os.getenv("SKYLOS_DEBUG"):
                        logger.error(traceback.format_exc())

        if progress_callback:
            progress_callback(0, 1, Path("PHASE: mark refs"))
        self._mark_refs(progress_callback=progress_callback)

        if progress_callback:
            progress_callback(0, 1, Path("PHASE: heuristics"))
        self._apply_heuristics()

        if progress_callback:
            progress_callback(0, 1, Path("PHASE: exports"))
        self._mark_exports()

        shown = 0

        def def_sort_key(d):
            return (d.type, d.name)

        for d in sorted(self.defs.values(), key=def_sort_key):
            if shown >= 50:
                break
            shown += 1

        unused = []
        for definition in self.defs.values():
            if (
                definition.references == 0
                and not definition.is_exported
                and definition.confidence > 0
                and definition.confidence >= thr
            ):
                unused.append(definition.to_dict())

        context_map = {}
        for name, d in self.defs.items():
            if d.type in ("class", "function", "method") and not name.startswith("_"):
                context_map[name] = {
                    "name": d.name,
                    "file": str(d.filename),
                    "line": d.line,
                    "type": d.type,
                }

        whitelisted = []
        for d in self.defs.values():
            reason = getattr(d, "skip_reason", None)
            if reason:
                whitelisted.append(
                    {
                        "name": d.simple_name,
                        "file": str(d.filename),
                        "line": d.line,
                        "reason": d.skip_reason,
                    }
                )

        result = {
            "definitions": context_map,
            "unused_functions": [],
            "unused_imports": [],
            "unused_classes": [],
            "unused_variables": [],
            "unused_parameters": [],
            "unused_files": [],
            "whitelisted": whitelisted,
            "analysis_summary": {
                "total_files": len(files),
                "excluded_folders": exclude_folders or [],
            },
        }

        if enable_secrets and all_secrets:
            result["secrets"] = all_secrets
            result["analysis_summary"]["secrets_count"] = len(all_secrets)

        if enable_danger and all_dangers:
            result["danger"] = all_dangers
            result["analysis_summary"]["danger_count"] = len(all_dangers)

        if enable_quality and all_quality:
            result["quality"] = all_quality
            result["analysis_summary"]["quality_count"] = len(all_quality)

        if empty_files:
            result["unused_files"] = empty_files
            result["analysis_summary"]["unused_files_count"] = len(empty_files)

        for u in unused:
            if u["type"] in ("function", "method"):
                result["unused_functions"].append(u)
            elif u["type"] == "import":
                result["unused_imports"].append(u)
            elif u["type"] == "class":
                result["unused_classes"].append(u)
            elif u["type"] == "variable":
                result["unused_variables"].append(u)
            elif u["type"] == "parameter":
                result["unused_parameters"].append(u)

        return json.dumps(result, indent=2)


def _is_truly_empty_or_docstring_only(tree):
    if not isinstance(tree, ast.Module):
        return False

    if not tree.body:
        return True

    if len(tree.body) != 1:
        return False

    only = tree.body[0]
    return (
        isinstance(only, ast.Expr)
        and isinstance(only.value, ast.Constant)
        and isinstance(only.value.value, str)
    )


def proc_file(file_or_args, mod=None, extra_visitors=None):
    if mod is None and isinstance(file_or_args, tuple):
        file, mod = file_or_args
    else:
        file = file_or_args

    cfg = load_config(file)

    if str(file).endswith((".ts", ".tsx")):
        ts_out = scan_typescript_file(file)
        if isinstance(ts_out, tuple) and len(ts_out) == 10:
            return (*ts_out, None, cfg)
        if isinstance(ts_out, tuple) and len(ts_out) == 11:
            return (*ts_out, cfg)
        if isinstance(ts_out, tuple) and len(ts_out) >= 12:
            return ts_out[:12]
        return (*ts_out, None, cfg)

    try:
        source = Path(file).read_text(encoding="utf-8")
        ignore_lines = get_all_ignore_lines(source)

        tree = ast.parse(source)

        empty_file_finding = None

        basename = Path(file).name
        skip_empty_report = basename in {"__init__.py", "__main__.py", "main.py"}

        if (
            _is_truly_empty_or_docstring_only(tree)
            and not skip_empty_report
            and "SKY-E002" not in cfg["ignore"]
        ):
            empty_file_finding = {
                "rule_id": "SKY-E002",
                "message": "Empty Python file (no code, or docstring-only)",
                "file": str(file),
                "line": 1,
                "severity": "LOW",
                "category": "DEAD_CODE",
            }

        from skylos.ast_mask import apply_body_mask, default_mask_spec_from_config

        mask = default_mask_spec_from_config(cfg)
        tree, masked = apply_body_mask(tree, mask)

        if masked and os.getenv("SKYLOS_DEBUG"):
            logger.info(f"{file}: masked {masked} bodies (skipped inner analysis)")

        q_rules = []
        if "SKY-Q301" not in cfg["ignore"]:
            q_rules.append(ComplexityRule(threshold=cfg["complexity"]))
        if "SKY-Q302" not in cfg["ignore"]:
            q_rules.append(NestingRule(threshold=cfg["nesting"]))
        if "SKY-C303" not in cfg["ignore"]:
            q_rules.append(ArgCountRule(max_args=cfg["max_args"]))
        if "SKY-C304" not in cfg["ignore"]:
            q_rules.append(FunctionLengthRule(max_lines=cfg["max_lines"]))

        if "SKY-L001" not in cfg["ignore"]:
            q_rules.append(MutableDefaultRule())
        if "SKY-L002" not in cfg["ignore"]:
            q_rules.append(BareExceptRule())
        if "SKY-L003" not in cfg["ignore"]:
            q_rules.append(DangerousComparisonRule())
        if "SKY-L004" not in cfg["ignore"]:
            q_rules.append(TryBlockPatternsRule(max_lines=15))

        if "SKY-U001" not in cfg["ignore"]:
            q_rules.append(UnreachableCodeRule())

        q_rules.append(PerformanceRule(ignore_list=cfg["ignore"]))

        linter_q = LinterVisitor(q_rules, str(file))
        linter_q.visit(tree)
        quality_findings = linter_q.findings

        d_rules = [DangerousCallsRule()]
        linter_d = LinterVisitor(d_rules, str(file))
        linter_d.visit(tree)
        danger_findings = linter_d.findings

        pro_findings = []
        if extra_visitors:
            for VisitorClass in extra_visitors:
                checker = VisitorClass(file, pro_findings)
                checker.visit(tree)

        tv = TestAwareVisitor(filename=file)
        tv.visit(tree)
        tv.ignore_lines = ignore_lines

        fv = FrameworkAwareVisitor(filename=file)
        fv.visit(tree)
        fv.finalize()
        v = Visitor(mod, file)
        v.visit(tree)

        fv.dataclass_fields = getattr(v, "dataclass_fields", set())
        fv.first_read_lineno = getattr(v, "first_read_lineno", {})
        fv.protocol_classes = getattr(v, "protocol_classes", set())
        fv.namedtuple_classes = getattr(v, "namedtuple_classes", set())
        fv.enum_classes = getattr(v, "enum_classes", set())
        fv.attrs_classes = getattr(v, "attrs_classes", set())
        fv.orm_model_classes = getattr(v, "orm_model_classes", set())
        fv.type_alias_names = getattr(v, "type_alias_names", set())
        fv.abc_classes = getattr(v, "abc_classes", set())
        fv.abstract_methods = getattr(v, "abstract_methods", {})
        fv.abc_implementers = getattr(v, "abc_implementers", {})
        fv.protocol_implementers = getattr(v, "protocol_implementers", {})
        fv.protocol_method_names = getattr(v, "protocol_method_names", {})

        return (
            v.defs,
            v.refs,
            v.dyn,
            v.exports,
            tv,
            fv,
            quality_findings,
            danger_findings,
            pro_findings,
            v.pattern_tracker,
            empty_file_finding,
            cfg,
        )

    except Exception as e:
        logger.error(f"{file}: {e}")
        if os.getenv("SKYLOS_DEBUG"):
            logger.error(traceback.format_exc())
        dummy_visitor = TestAwareVisitor(filename=file)
        dummy_visitor.ignore_lines = set()
        dummy_framework_visitor = FrameworkAwareVisitor(filename=file)
        return (
            [],
            [],
            set(),
            set(),
            dummy_visitor,
            dummy_framework_visitor,
            [],
            [],
            [],
            None,
            None,
            cfg
        )


def analyze(
    path,
    conf=60,
    exclude_folders=None,
    enable_secrets=False,
    enable_danger=False,
    enable_quality=False,
    extra_visitors=None,
    progress_callback=None,
):
    return Skylos().analyze(
        path,
        conf,
        exclude_folders,
        enable_secrets,
        enable_danger,
        enable_quality,
        extra_visitors,
        progress_callback,
    )


if __name__ == "__main__":
    enable_secrets = "--secrets" in sys.argv
    enable_danger = "--danger" in sys.argv
    enable_quality = "--quality" in sys.argv

    positional = [a for a in sys.argv[1:] if not a.startswith("--")]
    if not positional:
        print(
            "Usage: python Skylos.py <path> [confidence_threshold] [--secrets] [--danger] [--quality]"
        )
        sys.exit(2)
    p = positional[0]
    confidence = int(positional[1]) if len(positional) > 1 else 60

    result = analyze(
        p,
        confidence,
        enable_secrets=enable_secrets,
        enable_danger=enable_danger,
        enable_quality=enable_quality,
    )
    data = json.loads(result)
    print("\n Python Static Analysis Results")
    print("===================================\n")

    total_dead = 0
    for key, items in data.items():
        if key.startswith("unused_") and isinstance(items, list):
            total_dead += len(items)

    danger_count = (
        data.get("analysis_summary", {}).get("danger_count", 0) if enable_danger else 0
    )
    secrets_count = (
        data.get("analysis_summary", {}).get("secrets_count", 0)
        if enable_secrets
        else 0
    )

    print("Summary:")
    if data["unused_functions"]:
        print(f" * Unreachable functions: {len(data['unused_functions'])}")
    if data["unused_imports"]:
        print(f" * Unused imports: {len(data['unused_imports'])}")
    if data["unused_classes"]:
        print(f" * Unused classes: {len(data['unused_classes'])}")
    if data["unused_variables"]:
        print(f" * Unused variables: {len(data['unused_variables'])}")
    if data["unused_files"]:
        print(f" * Empty files: {len(data['unused_files'])}")
    if enable_danger:
        print(f" * Security issues: {danger_count}")
    if enable_secrets:
        print(f" * Secrets found: {secrets_count}")

    if data["unused_functions"]:
        print("\n - Unreachable Functions")
        print("=======================")
        for i, func in enumerate(data["unused_functions"], 1):
            print(f" {i}. {func['name']}")
            print(f"    └─ {func['file']}:{func['line']}")

    if data["unused_imports"]:
        print("\n - Unused Imports")
        print("================")
        for i, imp in enumerate(data["unused_imports"], 1):
            print(f" {i}. {imp['simple_name']}")
            print(f"    └─ {imp['file']}:{imp['line']}")

    if data["unused_classes"]:
        print("\n - Unused Classes")
        print("=================")
        for i, cls in enumerate(data["unused_classes"], 1):
            print(f" {i}. {cls['name']}")
            print(f"    └─ {cls['file']}:{cls['line']}")

    if data["unused_variables"]:
        print("\n - Unused Variables")
        print("==================")
        for i, var in enumerate(data["unused_variables"], 1):
            print(f" {i}. {var['name']}")
            print(f"    └─ {var['file']}:{var['line']}")

    if data["unused_files"]:
        print("\n - Empty Files")
        print("==============")
        for i, f in enumerate(data["unused_files"], 1):
            print(f" {i}. {f['file']}")
            print(f"    └─ Line {f['line']}")

    if enable_danger and data.get("danger"):
        print("\n - Security Issues")
        print("================")
        for i, f in enumerate(data["danger"], 1):
            print(
                f" {i}. {f['message']} [{f['rule_id']}] ({f['file']}:{f['line']}) Severity: {f['severity']}"
            )

    if enable_secrets and data.get("secrets"):
        print("\n - Secrets")
        print("==========")
        for i, s in enumerate(data["secrets"], 1):
            rid = s.get("rule_id", "SECRET")
            msg = s.get("message", "Potential secret")
            file = s.get("file")
            line = s.get("line", 1)
            sev = s.get("severity", "HIGH")
            print(f" {i}. {msg} [{rid}] ({file}:{line}) Severity: {sev}")

    print("\n" + "─" * 50)
    if enable_danger:
        print(
            f"Found {total_dead} dead code items and {danger_count} security flaws. Add this badge to your README:"
        )
    else:
        print(f"Found {total_dead} dead code items. Add this badge to your README:")
    print("```markdown")
    print(
        f"![Dead Code: {total_dead}](https://img.shields.io/badge/Dead_Code-{total_dead}_detected-orange?logo=codacy&logoColor=red)"
    )
    print("```")

    print("\nNext steps:")
    print("  * Use --interactive to select specific items to remove")
    print("  * Use --dry-run to preview changes before applying them")



import os
import importlib
import unittest
from unittest.mock import patch, MagicMock, mock_open

import skylos.api as api
from skylos.api import upload_report, extract_snippet


class TestSkylosApi(unittest.TestCase):
    @patch("subprocess.check_output")
    @patch("skylos.api.get_project_token")
    @patch("requests.post")
    def test_upload_report_success(self, mock_post, mock_token, mock_git):
        mock_token.return_value = "test_token_123"
        mock_git.side_effect = [b"mock_commit_hash\n", b"main\n", b"/mock/git/root\n"]

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"scanId": "scan_abc_789"}
        mock_post.return_value = mock_response

        dummy_results = {
            "danger": [
                {
                    "file": "app.py",
                    "line": 10,
                    "message": "High risk",
                    "rule_id": "SKY-D001",
                }
            ],
            "quality": [],
        }

        result = upload_report(dummy_results, is_forced=True)

        self.assertTrue(result["success"])
        self.assertEqual(result["scan_id"], "scan_abc_789")

        args, kwargs = mock_post.call_args
        payload = kwargs["json"]
        self.assertEqual(payload["commit_hash"], "mock_commit_hash")
        self.assertTrue(payload["is_forced"])
        self.assertEqual(payload["version"], "2.1.0")

    @patch("skylos.api.get_project_token")
    def test_upload_report_no_token(self, mock_token):
        mock_token.return_value = None
        result = upload_report({})
        self.assertFalse(result["success"])
        self.assertEqual(
            result["error"],
            "No token found. Run 'skylos sync connect' or set SKYLOS_TOKEN.",
        )

    @patch("subprocess.check_output")
    @patch("skylos.api.get_project_token")
    @patch("requests.post")
    def test_upload_report_retry_logic(self, mock_post, mock_token, mock_git):
        mock_token.return_value = "token"
        mock_git.return_value = b"test\n"

        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.text = "Internal Server Error"
        mock_post.return_value = mock_response

        result = upload_report({"danger": []})

        self.assertFalse(result["success"])
        self.assertEqual(mock_post.call_count, 3)
        self.assertIn("Server Error 500", result["error"])

    def test_extract_snippet_valid(self):
        content = "line1\nline2\nline3\nline4\nline5\n"
        with patch("builtins.open", mock_open(read_data=content)):
            snippet = extract_snippet("fake.py", 3, context=1)
            self.assertEqual(snippet, "line2\nline3\nline4")

    def test_extract_snippet_context_zero(self):
        content = "a\nb\nc\nd\n"
        with patch("builtins.open", mock_open(read_data=content)):
            snippet = extract_snippet("fake.py", 3, context=0)
            self.assertEqual(snippet, "c")

    def test_extract_snippet_missing_file_returns_none(self):
        with patch("builtins.open", side_effect=FileNotFoundError):
            snippet = extract_snippet("missing.py", 1, context=2)
            self.assertIsNone(snippet)

    @patch("skylos.api.get_project_token")
    @patch("skylos.api.get_git_info", return_value=("c", "b", "actor"))
    @patch("skylos.api.get_git_root", return_value=None)
    @patch("skylos.api.get_project_info")
    @patch("requests.post")
    def test_upload_report_whoami_failure_still_uploads(
        self, mock_post, mock_info, _, _mock_git_info, mock_token
    ):
        mock_token.return_value = "token"
        mock_info.return_value = None

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"scanId": "scan_ok"}
        mock_post.return_value = mock_response

        result = upload_report({"danger": []}, quiet=False)
        self.assertTrue(result["success"])
        self.assertEqual(result["scan_id"], "scan_ok")
        self.assertEqual(mock_post.call_count, 1)

    @patch("skylos.api.get_project_token")
    @patch("skylos.api.get_git_info", return_value=("c", "b", "actor"))
    @patch("skylos.api.get_git_root", return_value=None)
    @patch("requests.post")
    def test_upload_report_401_returns_invalid_token_error(
        self, mock_post, _, _mock_git_info, mock_token
    ):
        mock_token.return_value = "token"
        resp = MagicMock()
        resp.status_code = 401
        resp.text = "Unauthorized"
        mock_post.return_value = resp

        result = upload_report({"danger": []})
        self.assertFalse(result["success"])
        self.assertEqual(
            result["error"],
            "Invalid API token. Run 'skylos sync connect' to reconnect.",
        )
        self.assertEqual(mock_post.call_count, 1)

    @patch("skylos.api.get_project_token")
    @patch("skylos.api.get_git_info", return_value=("c", "b", "actor"))
    @patch("skylos.api.get_git_root", return_value=None)
    @patch("requests.post")
    def test_retry_returns_last_error_text(
        self, mock_post, _, _mock_git_info, mock_token
    ):
        mock_token.return_value = "token"

        r1 = MagicMock(status_code=500, text="E1")
        r2 = MagicMock(status_code=502, text="E2")
        r3 = MagicMock(status_code=503, text="E3")
        mock_post.side_effect = [r1, r2, r3]

        result = upload_report({"danger": []})
        self.assertFalse(result["success"])
        self.assertEqual(mock_post.call_count, 3)
        self.assertIn("Server Error 503", result["error"])
        self.assertIn("E3", result["error"])

    def test_base_url_api_suffix_endpoints(self):
        old = os.environ.get("SKYLOS_API_URL")
        try:
            os.environ["SKYLOS_API_URL"] = "https://example.com/api"
            importlib.reload(api)
            self.assertEqual(api.REPORT_URL, "https://example.com/api/report")
            self.assertEqual(api.WHOAMI_URL, "https://example.com/api/sync/whoami")
        finally:
            if old is None:
                os.environ.pop("SKYLOS_API_URL", None)
            else:
                os.environ["SKYLOS_API_URL"] = old
            importlib.reload(api)

    @patch("skylos.api.SarifExporter")
    @patch("skylos.api.get_project_token")
    @patch("skylos.api.get_git_info", return_value=("c", "b", "actor"))
    @patch("skylos.api.get_git_root", return_value=None)
    @patch("requests.post")
    def test_prepare_for_sarif_normalizes_missing_fields(
        self, mock_post, _root, _git, mock_token, mock_exporter
    ):
        mock_token.return_value = "token"

        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {"scanId": "scan_norm"}
        mock_post.return_value = resp

        mock_exporter.return_value.generate.return_value = {"version": "2.1.0"}

        result = upload_report({"danger": [{"message": "oops"}]}, quiet=True)
        self.assertTrue(result["success"])

        all_findings = mock_exporter.call_args[0][0]
        self.assertEqual(len(all_findings), 1)
        f = all_findings[0]
        self.assertEqual(f["rule_id"], "SKY-D000")
        self.assertEqual(f["line_number"], 1)
        self.assertEqual(f["file_path"], "unknown")
        self.assertEqual(f["category"], "SECURITY")
        self.assertEqual(f["message"], "oops")

    @patch("skylos.api.SarifExporter")
    @patch("skylos.api.get_project_token")
    @patch("skylos.api.get_git_info", return_value=("c", "b", "actor"))
    @patch("skylos.api.get_git_root", return_value="/mock/git/root")
    @patch("requests.post")
    def test_prepare_for_sarif_relpaths_when_git_root_present(
        self, mock_post, _root, _git, mock_token, mock_exporter
    ):
        mock_token.return_value = "token"

        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {"scanId": "scan_path"}
        mock_post.return_value = resp

        mock_exporter.return_value.generate.return_value = {"version": "2.1.0"}

        result = upload_report(
            {"danger": [{"file": "/mock/git/root/app.py", "line": 5, "message": "m"}]},
            quiet=True,
        )
        self.assertTrue(result["success"])

        all_findings = mock_exporter.call_args[0][0]
        f = all_findings[0]
        self.assertEqual(f["file_path"], "app.py")
        self.assertEqual(f["line_number"], 5)


if __name__ == "__main__":
    unittest.main()



import os
import requests
import subprocess
from skylos.credentials import get_key
from skylos.sarif_exporter import SarifExporter

BASE_URL = os.getenv("SKYLOS_API_URL", "https://skylos.dev").rstrip("/")

if BASE_URL.endswith("/api"):
    REPORT_URL = f"{BASE_URL}/report"
    WHOAMI_URL = f"{BASE_URL}/sync/whoami"
else:
    REPORT_URL = f"{BASE_URL}/api/report"
    WHOAMI_URL = f"{BASE_URL}/api/sync/whoami"


def get_project_token():
    return os.getenv("SKYLOS_TOKEN") or get_key("skylos_token")


def get_project_info(token):
    if not token:
        return None
    try:
        resp = requests.get(
            WHOAMI_URL,
            headers={"Authorization": f"Bearer {token}"},
            timeout=10,
        )
        if resp.status_code == 200:
            return resp.json()
    except Exception:
        pass
    return None


def get_git_root():
    try:
        return (
            subprocess.check_output(
                ["git", "rev-parse", "--show-toplevel"], stderr=subprocess.DEVNULL
            )
            .decode()
            .strip()
        )
    except Exception:
        return None


def get_git_info():
    try:
        commit = (
            subprocess.check_output(
                ["git", "rev-parse", "HEAD"], stderr=subprocess.DEVNULL
            )
            .decode()
            .strip()
        )
        branch = (
            subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"], stderr=subprocess.DEVNULL
            )
            .decode()
            .strip()
        )
        actor = os.getenv("GITHUB_ACTOR") or os.getenv("USER") or "unknown"
        return commit, branch, actor
    except Exception:
        return "unknown", "unknown", "unknown"


def extract_snippet(file_abs, line_number, context=3):
    if not file_abs:
        return None
    try:
        with open(file_abs, "r", encoding="utf-8", errors="ignore") as f:
            lines = f.readlines()
        start = max(0, line_number - 1 - context)
        end = min(len(lines), line_number + context)
        return "\n".join([line.rstrip("\n") for line in lines[start:end]])
    except Exception:
        return None


def upload_report(result_json, is_forced=False, quiet=False):
    token = get_project_token()
    if not token:
        return {
            "success": False,
            "error": "No token found. Run 'skylos sync connect' or set SKYLOS_TOKEN.",
        }

    if not quiet:
        info = get_project_info(token)
        if info and info.get("ok"):
            project_name = info.get("project", {}).get("name", "Unknown")
            print(f"Uploading to: {project_name}")

    commit, branch, actor = get_git_info()
    git_root = get_git_root()

    def prepare_for_sarif(items, category, default_rule_id=None):
        processed = []

        for item in items or []:
            finding = dict(item)

            rid = (
                finding.get("rule_id")
                or finding.get("rule")
                or finding.get("code")
                or finding.get("id")
                or default_rule_id
                or "UNKNOWN"
            )
            finding["rule_id"] = str(rid)

            raw_path = finding.get("file_path") or finding.get("file") or ""
            file_abs = os.path.abspath(raw_path) if raw_path else ""

            line_raw = finding.get("line_number") or finding.get("line") or 1
            try:
                line = int(line_raw)
            except Exception:
                line = 1
            if line < 1:
                line = 1
            finding["line_number"] = line

            if git_root and file_abs:
                try:
                    finding["file_path"] = os.path.relpath(file_abs, git_root).replace(
                        "\\", "/"
                    )
                except Exception:
                    finding["file_path"] = (
                        raw_path.replace("\\", "/") if raw_path else "unknown"
                    )
            else:
                finding["file_path"] = (
                    raw_path.replace("\\", "/") if raw_path else "unknown"
                )

            finding["category"] = category

            if not finding.get("message"):
                name = (
                    finding.get("name")
                    or finding.get("symbol")
                    or finding.get("function")
                    or ""
                )
                if category == "DEAD_CODE" and name:
                    finding["message"] = f"Dead code: {name}"
                else:
                    finding["message"] = (
                        finding.get("detail") or finding.get("msg") or "Issue"
                    )

            if file_abs and line:
                finding["snippet"] = (
                    finding.get("snippet") or extract_snippet(file_abs, line) or None
                )

            processed.append(finding)

        return processed

    all_findings = []

    all_findings.extend(
        prepare_for_sarif(result_json.get("danger", []), "SECURITY", "SKY-D000")
    )

    all_findings.extend(
        prepare_for_sarif(result_json.get("quality", []), "QUALITY", "SKY-Q000")
    )

    all_findings.extend(
        prepare_for_sarif(result_json.get("secrets", []), "SECRET", "SKY-S000")
    )

    all_findings.extend(
        prepare_for_sarif(
            result_json.get("unused_functions", []), "DEAD_CODE", "SKY-U001"
        )
    )
    all_findings.extend(
        prepare_for_sarif(
            result_json.get("unused_imports", []), "DEAD_CODE", "SKY-U002"
        )
    )
    all_findings.extend(
        prepare_for_sarif(
            result_json.get("unused_variables", []), "DEAD_CODE", "SKY-U003"
        )
    )
    all_findings.extend(
        prepare_for_sarif(
            result_json.get("unused_classes", []), "DEAD_CODE", "SKY-U004"
        )
    )

    exporter = SarifExporter(all_findings, tool_name="Skylos")
    payload = exporter.generate()

    payload.update(
        {
            "commit_hash": commit,
            "branch": branch,
            "actor": actor,
            "is_forced": bool(is_forced),
        }
    )

    last_err = None
    for _ in range(3):
        try:
            response = requests.post(
                REPORT_URL,
                json=payload,
                headers={"Authorization": f"Bearer {token}"},
                timeout=30,
            )
            if response.status_code in (200, 201):
                data = response.json()
                scan_id = data.get("scanId") or data.get("scan_id")
                passed = data.get("quality_gate_passed", True)

                if not quiet:
                    print(f"✓ Scan uploaded")
                    print(
                        f"{'PASS' if passed else 'FAIL'} Quality gate: {'PASSED' if passed else 'FAILED'}"
                    )
                    if scan_id:
                        print(f"\nView: {BASE_URL}/dashboard/scans/{scan_id}")

                return {
                    "success": True,
                    "scan_id": scan_id,
                    "quality_gate_passed": passed,
                }

            if response.status_code == 401:
                return {
                    "success": False,
                    "error": "Invalid API token. Run 'skylos sync connect' to reconnect.",
                }

            last_err = f"Server Error {response.status_code}: {response.text}"
        except Exception as e:
            last_err = f"Connection Error: {str(e)}"

    return {"success": False, "error": last_err or "Unknown error"}



import builtins
import os
import sys
import types
import pytest

from skylos.adapters.openai_adapter import OpenAIAdapter


class _FakeResponse:
    def __init__(self, output_text: str):
        self.output_text = output_text


class _FakeResponsesAPI:
    def __init__(self, *, should_raise: Exception | None = None):
        self.should_raise = should_raise
        self.last_kwargs = None

    def create(self, **kwargs):
        self.last_kwargs = kwargs
        if self.should_raise:
            raise self.should_raise
        return _FakeResponse("  hello world  ")


class _FakeClient:
    def __init__(self, api_key, responses_api: _FakeResponsesAPI):
        self.api_key = api_key
        self.responses = responses_api


def _install_fake_openai(monkeypatch, *, responses_api=None, capture=None):
    if responses_api is None:
        responses_api = _FakeResponsesAPI()

    def _OpenAI(api_key: str):
        if capture is not None:
            capture["api_key"] = api_key
        return _FakeClient(api_key=api_key, responses_api=responses_api)

    fake_openai = types.SimpleNamespace(OpenAI=_OpenAI)
    monkeypatch.setitem(sys.modules, "openai", fake_openai)
    return responses_api


def test_init_raises_if_openai_missing(monkeypatch):
    if "openai" in sys.modules:
        monkeypatch.delitem(sys.modules, "openai", raising=False)

    real_import = builtins.__import__

    def blocked_import(name, *args, **kwargs):
        if name == "openai":
            raise ImportError("nope")
        return real_import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", blocked_import)

    with pytest.raises(ImportError) as e:
        OpenAIAdapter(model="gpt-x", api_key="abc")
    assert "OpenAI not found" in str(e.value)


def test_init_raises_if_no_key(monkeypatch):
    _install_fake_openai(monkeypatch)
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)

    with pytest.raises(ValueError) as e:
        OpenAIAdapter(model="gpt-x", api_key=None)
    assert "No OpenAI API Key found" in str(e.value)


def test_init_uses_explicit_api_key(monkeypatch):
    cap = {}
    _install_fake_openai(monkeypatch, capture=cap)

    ad = OpenAIAdapter(model="gpt-x", api_key="MY_KEY")
    assert ad.client is not None
    assert cap["api_key"] == "MY_KEY"


def test_init_uses_env_key(monkeypatch):
    cap = {}
    _install_fake_openai(monkeypatch, capture=cap)

    monkeypatch.setenv("OPENAI_API_KEY", "ENV_KEY")
    ad = OpenAIAdapter(model="gpt-x", api_key=None)

    assert cap["api_key"] == "ENV_KEY"
    assert ad.client.api_key == "ENV_KEY"


def test_complete_success_calls_responses_create(monkeypatch):
    responses_api = _FakeResponsesAPI()
    _install_fake_openai(monkeypatch, responses_api=responses_api)

    ad = OpenAIAdapter(model="gpt-x", api_key="K")
    out = ad.complete("SYS", "USER")

    assert out == "hello world"  # stripped
    assert responses_api.last_kwargs == {
        "model": "gpt-x",
        "instructions": "SYS",
        "input": "USER",
    }


def test_complete_returns_error_string_on_exception(monkeypatch):
    responses_api = _FakeResponsesAPI(should_raise=RuntimeError("boom"))
    _install_fake_openai(monkeypatch, responses_api=responses_api)

    ad = OpenAIAdapter(model="gpt-x", api_key="K")
    out = ad.complete("SYS", "USER")

    assert out.startswith("OpenAI Error:")
    assert "boom" in out



import pytest
import tempfile
from pathlib import Path
from textwrap import dedent
from skylos.tracer import CallTracer


@pytest.fixture(scope="session")
def sample_project():
    with tempfile.TemporaryDirectory() as temp_dir:
        project_path = Path(temp_dir)

        create_sample_project(project_path)
        yield project_path


def create_sample_project(base_path: Path):
    (base_path / "app").mkdir()

    app_init = base_path / "app" / "__init__.py"
    app_init.write_text(
        dedent("""
        from .core import main_function
        from .utils import helper_function
        
        __all__ = ['main_function', 'helper_function']
    """)
    )

    core_py = base_path / "app" / "core.py"
    core_py.write_text(
        dedent("""
        import os
        import sys
        from typing import Dict, List  # List is unused here
        from collections import defaultdict
        
        def main_function():
            '''Main entry point - should not be flagged'''
            data = defaultdict(list)
            return process_data(data)
        
        def process_data(data: Dict):
            '''Used by main_function'''
            return len(data)
        
        def deprecated_function():
            '''This function is never called'''
            unused_var = "should be flagged"
            return unused_var
        
        def _private_helper():
            '''Private function, might be unused'''
            return "private"
    """)
    )

    utils_py = base_path / "app" / "utils.py"
    utils_py.write_text(
        dedent("""
        import json  # unused
        import re
        
        def helper_function():
            '''Exported function'''
            return validate_input("test")
        
        def validate_input(text: str):
            '''Used by helper_function'''
            return re.match(r"^[a-z]+$", text) is not None
        
        def unused_utility():
            '''Never called utility'''
            return "utility"
        
        class ConfigManager:
            '''Used class'''
            def __init__(self):
                self.config = {}
            
            def get(self, key):
                return self.config.get(key)
        
        class LegacyProcessor:
            '''Unused class'''
            def process(self, data):
                return data
    """)
    )

    tests_dir = base_path / "tests"
    tests_dir.mkdir()

    test_core = tests_dir / "test_core.py"
    test_core.write_text(
        dedent("""
        import unittest
        from app.core import main_function, process_data
        
        class TestCore(unittest.TestCase):
            def test_main_function(self):
                result = main_function()
                self.assertIsInstance(result, int)
            
            def test_process_data(self):
                data = {'key': ['value']}
                result = process_data(data)
                self.assertEqual(result, 1)
            
            def test_edge_case(self):
                '''Test methods should not be flagged'''
                pass
    """)
    )

    config_dir = base_path / "config"
    config_dir.mkdir()

    config_py = config_dir / "settings.py"
    config_py.write_text(
        dedent("""
        # Configuration file with potentially unused imports
        import os
        import logging
        
        DEBUG = True
        DATABASE_URL = os.getenv("DATABASE_URL", "sqlite:///app.db")
        
        def setup_logging():
            '''Might be called externally'''
            logging.basicConfig(level=logging.INFO)
    """)
    )

    pycache_dir = base_path / "__pycache__"
    pycache_dir.mkdir()

    cache_file = pycache_dir / "cached.pyc"
    cache_file.write_bytes(b"fake compiled python")


@pytest.fixture
def simple_python_file():
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as f:
        f.write(
            dedent("""
            import os  # unused
            import sys
            
            def used_function():
                print("Hello", file=sys.stderr)
            
            def unused_function():
                return "never called"
            
            if __name__ == "__main__":
                used_function()
        """)
        )
        f.flush()
        yield Path(f.name)
        Path(f.name).unlink()


@pytest.fixture
def empty_project():
    with tempfile.TemporaryDirectory() as temp_dir:
        yield Path(temp_dir)


@pytest.fixture
def project_with_syntax_error():
    with tempfile.TemporaryDirectory() as temp_dir:
        project_path = Path(temp_dir)

        valid_py = project_path / "valid.py"
        valid_py.write_text("def valid_function():\n    return True\n")

        invalid_py = project_path / "invalid.py"
        invalid_py.write_text(
            "def invalid_function(\n    # Missing closing parenthesis\n    return False\n"
        )

        yield project_path


@pytest.fixture(autouse=True)
def cleanup_temp_files():
    """auto cleanup any temporary files after each test"""
    yield


pytest_plugins = []


def pytest_configure(config):
    config.addinivalue_line("markers", "integration: mark test as an integration test")
    config.addinivalue_line("markers", "unit: mark test as a unit test")
    config.addinivalue_line("markers", "slow: mark test as slow running")
    config.addinivalue_line("markers", "cli: mark test as CLI interface test")


@pytest.fixture
def mock_git_repo():
    with tempfile.TemporaryDirectory() as temp_dir:
        repo_path = Path(temp_dir)

        try:
            import subprocess

            subprocess.run(
                ["git", "init"], cwd=repo_path, check=True, capture_output=True
            )
            subprocess.run(
                ["git", "config", "user.email", "test@example.com"],
                cwd=repo_path,
                check=True,
            )
            subprocess.run(
                ["git", "config", "user.name", "Test User"], cwd=repo_path, check=True
            )

            test_file = repo_path / "test.py"
            test_file.write_text("def test(): pass\n")
            subprocess.run(["git", "add", "."], cwd=repo_path, check=True)
            subprocess.run(
                ["git", "commit", "-m", "Initial commit"], cwd=repo_path, check=True
            )
        except (subprocess.CalledProcessError, FileNotFoundError):
            pass

        yield repo_path


_tracer = None


def pytest_addoption(parser):
    parser.addoption(
        "--skylos-trace",
        action="store_true",
        default=False,
        help="Enable Skylos call tracing to capture dynamic function calls",
    )
    parser.addoption(
        "--skylos-trace-include",
        action="store",
        default=None,
        help="Comma-separated path patterns to include (e.g., 'skylos/,myproject/')",
    )
    parser.addoption(
        "--skylos-trace-output",
        action="store",
        default=".skylos_trace",
        help="Output file for trace data (default: .skylos_trace)",
    )


def pytest_configure(config):
    global _tracer

    if config.getoption("--skylos-trace"):
        include = config.getoption("--skylos-trace-include")
        include_patterns = include.split(",") if include else None

        _tracer = CallTracer(include_patterns=include_patterns)
        _tracer.start()
        print("\nSkylos call tracing enabled")


def pytest_unconfigure(config):
    global _tracer

    if _tracer is not None:
        _tracer.stop()
        output_path = config.getoption("--skylos-trace-output", ".skylos_trace")
        _tracer.save(output_path)

        stats = _tracer.get_stats()
        print(
            f"\nSkylos trace: {stats['unique_functions']} functions across {stats['files_traced']} files"
        )



import builtins
import sys
import types
import pytest

from skylos.adapters.anthropic_adapter import AnthropicAdapter


class _FakeMsg:
    def __init__(self, text: str):
        self.text = text


class _FakeResponse:
    def __init__(self, text: str):
        self.content = [_FakeMsg(text)]


class _FakeMessagesAPI:
    def __init__(self, *, should_raise: Exception | None = None, text="ok"):
        self.should_raise = should_raise
        self.text = text
        self.last_kwargs = None

    def create(self, **kwargs):
        self.last_kwargs = kwargs
        if self.should_raise:
            raise self.should_raise
        return _FakeResponse(self.text)


class _FakeClient:
    def __init__(self, api_key: str, messages_api: _FakeMessagesAPI):
        self.api_key = api_key
        self.messages = messages_api


def _install_fake_anthropic(monkeypatch, *, messages_api=None, capture=None):
    if messages_api is None:
        messages_api = _FakeMessagesAPI()

    def _Anthropic(api_key: str):
        if capture is not None:
            capture["api_key"] = api_key
        return _FakeClient(api_key=api_key, messages_api=messages_api)

    fake_anthropic = types.SimpleNamespace(Anthropic=_Anthropic)
    monkeypatch.setitem(sys.modules, "anthropic", fake_anthropic)
    return messages_api


def test_init_raises_if_anthropic_missing(monkeypatch):
    if "anthropic" in sys.modules:
        monkeypatch.delitem(sys.modules, "anthropic", raising=False)

    real_import = builtins.__import__

    def blocked_import(name, *args, **kwargs):
        if name == "anthropic":
            raise ImportError("nope")
        return real_import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", blocked_import)

    with pytest.raises(ImportError) as e:
        AnthropicAdapter(model="claude-x", api_key="abc")
    assert "Anthropic SDK not found" in str(e.value)


def test_init_raises_if_no_key(monkeypatch):
    _install_fake_anthropic(monkeypatch)
    monkeypatch.delenv("ANTHROPIC_API_KEY", raising=False)

    with pytest.raises(ValueError) as e:
        AnthropicAdapter(model="claude-x", api_key=None)
    assert "No Anthropic API Key found" in str(e.value)


def test_init_uses_explicit_api_key(monkeypatch):
    cap = {}
    _install_fake_anthropic(monkeypatch, capture=cap)

    ad = AnthropicAdapter(model="claude-x", api_key="MY_KEY")
    assert ad.client is not None
    assert cap["api_key"] == "MY_KEY"


def test_init_uses_env_key(monkeypatch):
    cap = {}
    _install_fake_anthropic(monkeypatch, capture=cap)

    monkeypatch.setenv("ANTHROPIC_API_KEY", "ENV_KEY")
    ad = AnthropicAdapter(model="claude-x", api_key=None)

    assert cap["api_key"] == "ENV_KEY"
    assert ad.client.api_key == "ENV_KEY"


def test_complete_success_calls_messages_create(monkeypatch):
    messages_api = _FakeMessagesAPI(text="hello from claude")
    _install_fake_anthropic(monkeypatch, messages_api=messages_api)

    ad = AnthropicAdapter(model="claude-x", api_key="K")
    out = ad.complete("SYS", "USER")

    assert out == "hello from claude"
    assert messages_api.last_kwargs == {
        "model": "claude-x",
        "max_tokens": 4096,
        "system": "SYS",
        "messages": [{"role": "user", "content": "USER"}],
        "temperature": 0.2,
    }


def test_complete_returns_error_string_on_exception(monkeypatch):
    messages_api = _FakeMessagesAPI(should_raise=RuntimeError("boom"))
    _install_fake_anthropic(monkeypatch, messages_api=messages_api)

    ad = AnthropicAdapter(model="claude-x", api_key="K")
    out = ad.complete("SYS", "USER")

    assert out.startswith("Anthropic Error:")
    assert "boom" in out
